"""
Simple IO helpers
"""

from __future__ import annotations
from typing import BinaryIO


def write_int(f: BinaryIO, data: int):
    f.write(data.to_bytes(4, "little"))
    f.flush()


def read_int(f: BinaryIO) -> int:
    return int.from_bytes(f.read(4), "little")


def write_str(f: BinaryIO, data: str):
    write_bytes(f, data.encode("utf8"))


def read_str(f: BinaryIO) -> str:
    return read_bytes(f).decode("utf8")


def write_bytes(f: BinaryIO, data: bytes):
    f.write(len(data).to_bytes(4, "little"))
    f.write(data)
    f.flush()


def read_bytes(f: BinaryIO) -> bytes:
    data_len = read_int(f)
    return f.read(data_len)


def write_str_array(f: BinaryIO, data: list[str]):
    write_int(f, len(data))
    for item in data:
        write_str(f, item)


def read_str_array(f: BinaryIO) -> list[str]:
    data_len = read_int(f)
    return [read_str(f) for _ in range(data_len)]


def read_expected(f: BinaryIO, expected: bytes):
    res = f.read(len(expected))
    assert res == expected, f"Expected {expected!r}, got {res!r}"



"""
fork server method
"""

from typing import List, Tuple
import os
import socket
import sys
import fcntl
import termios
import array
import select
from . import _io, utils


def child_main(*, sock: socket.socket):
    """child for fork-server"""
    p2c_r, c2p_w = sock.makefile("rb"), sock.makefile("wb")
    _io.write_str(c2p_w, os.getcwd())
    _io.write_str_array(c2p_w, sys.argv)

    print("Open new PTY", file=sys.stderr)
    master_fd, slave_fd = os.openpty()
    print("Send PTY fd to server", file=sys.stderr)
    _send_fds(sock, b"pty", [slave_fd])
    print("Wait for server to be ready", file=sys.stderr)
    _io.read_expected(p2c_r, b"ok")
    print("Entering PTY proxy loop", file=sys.stderr)
    os.close(slave_fd)

    tcattr = termios.tcgetattr(0)
    orig_tcattr = tcattr.copy()
    try:
        # TTY raw mode (cfmakeraw)
        tcattr[0] &= ~(termios.IGNBRK | termios.BRKINT | termios.IGNPAR | termios.PARMRK | termios.INPCK |
                       termios.ISTRIP | termios.INLCR | termios.IGNCR | termios.ICRNL | termios.IXON |
                       termios.IXANY | termios.IXOFF)
        tcattr[1] &= ~termios.OPOST
        tcattr[2] &= ~(termios.PARENB | termios.CSIZE)
        tcattr[2] |= termios.CS8
        tcattr[3] &= ~(termios.ECHO | termios.ECHOE | termios.ECHOK | termios.ECHONL | termios.ICANON |
                       termios.IEXTEN | termios.ISIG | termios.NOFLSH | termios.TOSTOP)
        tcattr[6][termios.VMIN] = 1
        tcattr[6][termios.VTIME] = 0
        termios.tcsetattr(0, termios.TCSANOW, tcattr)

        # PTY proxy
        pty = os.fdopen(master_fd, "wb")
        stdout = os.fdopen(1, "wb")
        while True:
            rl, _, _ = select.select([0, master_fd], [], [])
            if master_fd in rl:
                try:
                    buf = os.read(master_fd, 4096)
                except OSError:  # Input/output error, when closed
                    break
                if not buf:
                    break
                stdout.write(buf)
                stdout.flush()
            if 0 in rl:
                buf = os.read(0, 4096)
                if not buf:
                    break
                pty.write(buf)
                pty.flush()
    finally:
        termios.tcsetattr(0, termios.TCSANOW, orig_tcattr)
        sock.close()
    sys.exit()


def server_main(*, sock: socket.socket, modules: List[str], file_prefix: str):
    """server for fork-server"""
    server_preload(modules=modules)

    if os.fork() != 0:
        # In parent.
        # This was started when the bundled app was started for the first time,
        # thus we were given some arguments which also should be handled now.
        utils.child_run(sys.argv)
        # Leave the child alone. See below.
        return

    # See docs/pty-details.md for some background.
    # In child. Further logic here.
    # This is such that the child is not a process group leader,
    # such that os.setsid() works below.
    # Detach the TTY now.
    fd = os.open("/dev/tty", os.O_RDWR | os.O_NOCTTY)
    fcntl.ioctl(fd, termios.TIOCNOTTY)
    os.close(0)
    os.dup2(os.open(file_prefix + ".stdout", os.O_CREAT | os.O_APPEND | os.O_WRONLY), 1)
    os.dup2(os.open(file_prefix + ".stderr", os.O_CREAT | os.O_APPEND | os.O_WRONLY), 2)

    # now wait for other childs
    while True:
        conn, _ = sock.accept()
        server_handle_child(conn)
        conn.close()


def server_preload(*, modules: List[str]):
    """
    Preload
    """
    import importlib

    for mod_name in modules:
        print("Import module:", mod_name)
        importlib.import_module(mod_name)


def server_handle_child(conn: socket.socket):
    """handle child for fork-server"""
    c2p_r, p2c_w = conn.makefile("rb"), conn.makefile("wb")
    cwd = _io.read_str(c2p_r)
    args = _io.read_str_array(c2p_r)
    print("Handle child:", args, file=sys.stderr)

    msg, (slave_fd,) = _recv_fds(conn, msglen=3, maxfds=1)
    assert msg == b"pty"
    print("Got PTY fd from client", file=sys.stderr)
    os.set_inheritable(slave_fd, True)
    p2c_w.write(b"ok")
    p2c_w.flush()

    pid = os.fork()
    if pid == 0:  # child:
        os.setsid()
        fcntl.ioctl(slave_fd, termios.TIOCSCTTY, 1)
        os.dup2(slave_fd, 0)
        os.dup2(slave_fd, 1)
        os.dup2(slave_fd, 2)
        if slave_fd > 2:
            os.close(slave_fd)
        os.chdir(cwd)
        utils.child_run(args)
        sys.exit(0)

    # parent
    os.close(slave_fd)
    # continue, wait for potential other childs


def _send_fds(sock: socket.socket, msg: bytes, fds: List[int]):
    return sock.sendmsg([msg], [(socket.SOL_SOCKET, socket.SCM_RIGHTS, array.array("i", fds))])


def _recv_fds(sock: socket.socket, msglen: int, maxfds: int) -> Tuple[bytes, List[int]]:
    fds = array.array("i")   # Array of ints
    msg, ancdata, flags, addr = sock.recvmsg(msglen, socket.CMSG_LEN(maxfds * fds.itemsize))
    for cmsg_level, cmsg_type, cmsg_data in ancdata:
        if cmsg_level == socket.SOL_SOCKET and cmsg_type == socket.SCM_RIGHTS:
            # Append data, ignoring any truncated integers at the end.
            fds.frombytes(cmsg_data[:len(cmsg_data) - (len(cmsg_data) % fds.itemsize)])
    return msg, list(fds)



"""
Simple CRIU wrapper
"""

import os
import subprocess
from . import utils


def dump(checkpoint_path: str, pid: int):
    """
    Dump
    """
    os.makedirs(checkpoint_path, exist_ok=True)
    subprocess.check_call(["sudo", "criu", "dump", "-t", str(pid), "-D", checkpoint_path, "--shell-job"])


def restore(checkpoint_path: str, *, p2c_r_fd: int, old_pipe_ino: int):
    """
    restore. does not return when successful
    """
    criu_bin = utils.which("criu")
    if not criu_bin:
        raise Exception("Cannot find criu binary in PATH")
    os.execl(
        criu_bin,
        criu_bin, "restore",
        "-D", checkpoint_path,
        "--inherit-fd", f"fd[{p2c_r_fd}]:pipe:[{old_pipe_ino}]")



"""
Startup helper for preloaded bundle
"""

from __future__ import annotations
from typing import BinaryIO, List
import os
import sys
import socket

from . import _io, utils


def startup_after_criu_dump(*, p2c_r: BinaryIO):
    """
    Main entry, right after the dump.
    I.e. this is the context after restore, with the preloaded modules.
    """

    args = _io.read_str_array(p2c_r)
    utils.child_run(args)


def startup_restore_criu(*, checkpoint_path: str, p2c_r_fd: int, old_pipe_ino: int):
    """main entry, prepare restore"""
    from . import criu

    p2c_r_new_fd, p2c_w_fd = os.pipe()
    os.dup2(p2c_r_new_fd, p2c_r_fd, inheritable=True)
    os.close(p2c_r_new_fd)

    child_pid = os.fork()
    if child_pid == 0:  # child
        criu.restore(checkpoint_path, p2c_r_fd=p2c_r_fd, old_pipe_ino=old_pipe_ino)
        raise Exception("Should not return from criu.restore()")

    # parent
    os.close(p2c_r_fd)
    p2c_w = os.fdopen(p2c_w_fd, "wb")
    _io.write_bytes(p2c_w, b"startup_after_dump")
    _io.write_str_array(p2c_w, sys.argv)
    p, res = os.waitpid(child_pid, 0)
    assert p == child_pid, f"exit code {res} from pid {p}, child pip {child_pid}"
    sys.exit(res)


def startup_via_fork_server(*, modules: List[str]):
    """fork server method"""
    from . import fork_server
    import platform
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.set_inheritable(True)
    file_prefix = sys.argv[0] + ".server." + platform.node()
    sock_name = file_prefix + ".socket"
    if os.path.exists(sock_name):
        try:
            sock.connect(sock_name)
        except socket.error as exc:
            print("Existing socket but can not connect:", exc, file=sys.stderr)
            os.unlink(sock_name)
        else:
            print("Existing socket, connected", file=sys.stderr)
            fork_server.child_main(sock=sock)
            return

    sock.bind(sock_name)
    sock.listen()
    fork_server.server_main(sock=sock, modules=modules, file_prefix=file_prefix)



"""
Generic utils
"""

from typing import Optional, List
import sys
import os


def child_run(args: List[str]):
    """run passed args"""
    import runpy
    import types
    import builtins as builtin_mod

    # make sure we have a fresh __main__
    sys.modules.pop("__main__", None)
    # Partly copied from IPython.
    new_main_module = types.ModuleType(
        "__main__", doc="Automatically created __main__ module by python-preloaded")
    # We must ensure that __builtin__ (without the final 's') is always
    # available and pointing to the __builtin__ *module*.  For more details:
    # http://mail.python.org/pipermail/python-dev/2001-April/014068.html
    new_main_module.__dict__.setdefault('__builtin__', builtin_mod)
    new_main_module.__dict__.setdefault('__builtins__', builtin_mod)
    sys.modules["__main__"] = new_main_module

    # args[0] should be the bundled Python script
    sys.argv = args[1:]
    if not sys.argv:
        import code
        code.interact()
    elif sys.argv[0] in ["-h", "--help"]:
        print(f"Usage: {args[0]} [-c cmd | -m mod | file | -] [arg] ...")
    elif sys.argv[0] == "-m":
        sys.argv = sys.argv[1:]
        assert sys.argv, "missing module name after -m"
        runpy.run_module(sys.argv[0], run_name="__main__", alter_sys=True)
    elif sys.argv[0] == "-c":
        assert len(sys.argv) == 2, f"expect exactly one arg after -c but got args {sys.argv}"
        exec(sys.argv[1])
    else:
        script_path = os.path.realpath(sys.argv[0])
        sys.path.insert(0, os.path.dirname(script_path))
        runpy.run_path(script_path, run_name="__main__")


def which(program: str) -> Optional[str]:
    """
    Finds `program` in the directories of the PATH env var.

    :param program: e.g. "criu"
    :return: full path, e.g. "/usr/bin/criu", or None
    """
    if "/" in program:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(":"):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def is_exe(path: str) -> bool:
    """
    Checks if `path` is an executable file.
    """
    return os.path.isfile(path) and os.access(path, os.X_OK)
